# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

from .entity_with_metadata import EntityWithMetadata
from ..block import BlockMixin
from .exceptions import exceptions
from .group import Group
from .data_array import DataArray
from .multi_tag import MultiTag
from .tag import Tag
from .source import Source
from . import util


class Block(EntityWithMetadata, BlockMixin):

    def __init__(self, nixparent, h5group):
        super(Block, self).__init__(nixparent, h5group)
        # TODO: Validation for containers

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_):
        newentity = super(Block, cls)._create_new(nixparent, h5parent,
                                                  name, type_)
        return newentity

    # DataArray
    def _create_data_array(self, name, type_, data_type, shape):
        util.check_entity_name_and_type(name, type_)
        data_arrays = self._h5group.open_group("data_arrays")
        if name in data_arrays:
            raise exceptions.DuplicateName("create_data_array")
        da = DataArray._create_new(self, data_arrays, name, type_,
                                   data_type, shape)
        return da

    def _get_data_array_by_id(self, id_or_name):
        data_arrays = self._h5group.open_group("data_arrays")
        return DataArray(self, data_arrays.get_by_id_or_name(id_or_name))

    def _get_data_array_by_pos(self, pos):
        data_arrays = self._h5group.open_group("data_arrays")
        return DataArray(self, data_arrays.get_by_pos(pos))

    def _delete_data_array_by_id(self, id_):
        data_arrays = self._h5group.open_group("data_arrays")
        data_arrays.delete(id_)

    def _data_array_count(self):
        data_arrays = self._h5group.open_group("data_arrays")
        return len(data_arrays)

    # MultiTag
    def create_multi_tag(self, name, type_, positions):
        """
        Create a new multi tag for this block.

        :param name: The name of the tag to create.
        :type name: str
        :param type_: The type of tag.
        :type type_: str
        :param positions: A data array defining all positions of the tag.
        :type positions: DataArray

        :returns: The newly created tag.
        :rtype: MultiTag
        """
        util.check_entity_name_and_type(name, type_)
        util.check_entity_input(positions)
        if not isinstance(positions, DataArray):
            raise TypeError("DataArray expected for 'positions'")
        multi_tags = self._h5group.open_group("multi_tags")
        if name in multi_tags:
            raise exceptions.DuplicateName("create_multi_tag")
        mtag = MultiTag._create_new(self, multi_tags, name, type_, positions)
        return mtag

    def _get_multi_tag_by_id(self, id_or_name):
        multi_tags = self._h5group.open_group("multi_tags")
        return MultiTag(self, multi_tags.get_by_id_or_name(id_or_name))

    def _get_multi_tag_by_pos(self, pos):
        multi_tags = self._h5group.open_group("multi_tags")
        return MultiTag(self, multi_tags.get_by_pos(pos))

    def _delete_multi_tag_by_id(self, id_):
        multi_tags = self._h5group.open_group("multi_tags")
        multi_tags.delete(id_)

    def _multi_tag_count(self):
        multi_tags = self._h5group.open_group("multi_tags")
        return len(multi_tags)

    # Tag
    def create_tag(self, name, type_, position):
        """
        Create a new tag for this block.

        :param name: The name of the tag to create.
        :type name: str
        :param type_: The type of tag.
        :type type_: str
        :param position: Coordinates of the start position
                         in units of the respective data dimension.

        :returns: The newly created tag.
        :rtype: Tag
        """
        util.check_entity_name_and_type(name, type_)
        tags = self._h5group.open_group("tags")
        if name in tags:
            raise exceptions.DuplicateName("create_tag")
        tag = Tag._create_new(self, tags, name, type_, position)
        return tag

    def _get_tag_by_id(self, id_or_name):
        tags = self._h5group.open_group("tags")
        return Tag(self, tags.get_by_id_or_name(id_or_name))

    def _get_tag_by_pos(self, pos):
        tags = self._h5group.open_group("tags")
        return Tag(self, tags.get_by_pos(pos))

    def _delete_tag_by_id(self, id_):
        tags = self._h5group.open_group("tags")
        tags.delete(id_)

    def _tag_count(self):
        tags = self._h5group.open_group("tags")
        return len(tags)

    # Source
    def create_source(self, name, type_):
        """
        Create a new source on this block.

        :param name: The name of the source to create.
        :type name: str
        :param type_: The type of the source.
        :type type_: str

        :returns: The newly created source.
        :rtype: Source
        """
        util.check_entity_name_and_type(name, type_)
        sources = self._h5group.open_group("sources")
        if name in sources:
            raise exceptions.DuplicateName("create_source")
        src = Source._create_new(self, sources, name, type_)
        return src

    def _get_source_by_id(self, id_or_name):
        sources = self._h5group.open_group("sources")
        return Source(self, sources.get_by_id_or_name(id_or_name))

    def _get_source_by_pos(self, pos):
        sources = self._h5group.open_group("sources")
        return Source(self, sources.get_by_pos(pos))

    def _delete_source_by_id(self, id_):
        sources = self._h5group.open_group("sources")
        sources.delete(id_)

    def _source_count(self):
        sources = self._h5group.open_group("sources")
        return len(sources)

    # Group
    def create_group(self, name, type_):
        """
        Create a new group on this block.

        :param name: The name of the group to create.
        :type name: str
        :param type_: The type of the group.
        :type type_: str

        :returns: The newly created group.
        :rtype: Group
        """
        util.check_entity_name_and_type(name, type_)
        groups = self._h5group.open_group("groups")
        if name in groups:
            raise exceptions.DuplicateName("open_group")
        grp = Group._create_new(self, groups, name, type_)
        return grp

    def _get_group_by_id(self, id_or_name):
        groups = self._h5group.open_group("groups")
        return Group(self, groups.get_by_id_or_name(id_or_name))

    def _get_group_by_pos(self, pos):
        groups = self._h5group.open_group("groups")
        return Group(self, groups.get_by_pos(pos))

    def _delete_group_by_id(self, id_):
        groups = self._h5group.open_group("groups")
        groups.delete(id_)

    def _group_count(self):
        groups = self._h5group.open_group("groups")
        return len(groups)



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)


class PropertyMixin(object):

    def __eq__(self, other):
        if hasattr(other, "id"):
            return self.id == other.id
        else:
            return False

    def __hash__(self):
        """
        overwriting method __eq__ blocks inheritance of __hash__ in Python 3
        hash has to be either explicitly inherited from parent class,
        implemented or escaped
        """
        return hash(self.id)



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import nixio.util.find as finders
from nixio.block import SourceProxyList

try:
    from sys import maxint
except:
    from sys import maxsize as maxint


class SourceMixin(object):

    def find_sources(self, filtr=lambda _: True, limit=None):
        """
        Get all child sources of this source recursively.

        This method traverses the tree of all sources. The traversal
        is accomplished via breadth first and can be limited in depth. On each
        node or source a filter is applied. If the filter returns true the
        respective source will be added to the result list.
        By default a filter is used that accepts all sources.

        :param filtr: A filter function
        :type filtr:  function
        :param limit: The maximum depth of traversal
        :type limit:  int

        :returns: A list containing the matching sources.
        :rtype: list of Source
        """
        if limit is None:
            limit = maxint
        return finders._find_sources(self, filtr, limit)

    @property
    def sources(self):
        """
        A property containing all sources of a block. Sources can be obtained
        via their index or by their id. Sources can be deleted from the list.
        Adding sources is done using the Blocks create_source method.
        This is a read only attribute.

        :type: ProxyList of Source entities.
        """
        if not hasattr(self, "_sources"):
            setattr(self, "_sources", SourceProxyList(self))
        return self._sources

    def __eq__(self, other):
        if hasattr(other, "id"):
            return self.id == other.id
        else:
            return False

    def __hash__(self):
        """
        overwriting method __eq__ blocks inheritance of __hash__ in Python 3
        hash has to be either explicitly inherited from parent class,
        implemented or escaped
        """
        return hash(self.id)



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

import h5py
import numpy as np

from .h5dataset import H5DataSet
from ..value import DataType
from .block import Block
from .section import Section

from . import util
from .exceptions import InvalidEntity


class H5Group(object):

    def __init__(self, parent, name, create=False):
        self._parent = parent
        self.name = name
        self.group = None
        if create or name in self._parent:
            self._create_h5obj()

    def _create_h5obj(self):
        if self.name in self._parent:
            self.group = self._parent[self.name]
        else:
            gcpl = h5py.h5p.create(h5py.h5p.GROUP_CREATE)
            flags = h5py.h5p.CRT_ORDER_TRACKED | h5py.h5p.CRT_ORDER_INDEXED
            gcpl.set_link_creation_order(flags)
            name = self.name.encode("utf-8")
            gid = h5py.h5g.create(self._parent.id, name, gcpl=gcpl)
            self.group = h5py.Group(gid)

    def create_link(self, target, name):
        self._create_h5obj()
        if name in self.group:
            del self.group[name]
        self.group[name] = target._h5group.group

    @classmethod
    def create_from_h5obj(cls, h5obj):
        parent = h5obj.parent
        name = h5obj.name.split("/")[-1]
        if isinstance(h5obj, h5py.Group):
            return cls(parent, name)
        elif isinstance(h5obj, h5py.Dataset):
            return H5DataSet(parent, name)
        else:
            raise ValueError("Invalid object: "
                             "{} must be either h5py.Group of h5py.Dataset.")

    def open_group(self, name, create=False):
        """
        Returns a new H5Group with the given name contained in the current
        group.  If the current group does not exist in the file,
        it is automatically created.

        :param name: the name of the group
        :param create: creates the child group in the file if it does not exist
        :return: a new H5Group object
        """
        self._create_h5obj()
        return H5Group(self.group, name, create)

    def create_dataset(self, name, shape, dtype):
        """
        Creates a dataset object under the current group with a given name,
        shape, and type.

        :param name: the name of the dataset
        :param shape: tuple representing the shape of the dataset
        :param dtype: the type of the data for this dataset (DataType)
        :return: a new H5DataSet object
        """
        self._create_h5obj()
        return H5DataSet(self.group, name, dtype, shape)

    def get_dataset(self, name):
        """
        Returns a contained H5DataSet object.

        :param name: name of the dataset
        :return: H5DataSet object
        """
        notfound = KeyError("No DataSet named {} found.")
        if self.group is None:
            raise notfound
        if name in self.group:
            dset = self.group[name]
            return H5DataSet.create_from_h5obj(dset)
        else:
            raise notfound

    def write_data(self, name, data, dtype=None):
        """
        Writes the data to a Dataset contained in the group with the
        given name.  Creates the Dataset if necessary.

        :param name: name of the Dataset object
        :param data: the data to write
        :param dtype: optionally specify the data type, otherwise it will be
        automatically determined by the data
        """
        shape = np.shape(data)
        if self.has_data(name):
            dset = self.get_dataset(name)
            dset.shape = shape
        else:
            if dtype is None:
                dtype = DataType.get_dtype(data[0])
            dset = self.create_dataset(name, shape, dtype)

        dset.write_data(data)

    def get_data(self, name):
        """
        Returns the data contained in the dataset identified by 'name', or an
        empty list if a dataset of that name does not exist in the Group.

        :param name: The name of the dataset
        :return: The data contained in the dataset as a numpy array or None
        """
        if name not in self.group:
            return []

        dset = self.group[name]
        # TODO: Error if dset is Group?
        return dset[:]

    def has_data(self, name):
        """
        Return True if the Group contains a Dataset object with the given name.

        :param name: name of Dataset
        :return: True if Dataset exists in Group, False if it does not exist,
        or exists and is not a Dataset
        """
        if self.group.get(name, getclass=True) == h5py.Dataset:
            return True
        else:
            return False

    def has_by_id(self, id_or_name):
        if not self.group:
            return False
        if util.is_uuid(id_or_name):
            for item in self:
                if item.get_attr("entity_id") == id_or_name:
                    return True
            else:
                return False
        else:
            return id_or_name in self.group

    def get_by_id_or_name(self, id_or_name):
        if util.is_uuid(id_or_name):
            return self.get_by_id(id_or_name)
        else:
            return self.get_by_name(id_or_name)

    def get_by_name(self, name):
        if self.group and name in self.group:
            return self.create_from_h5obj(self.group[name])
        else:
            raise ValueError("No item with name {} found in {}".format(
                name, self.group.name
            ))

    def get_by_id(self, id_):
        if self.group:
            for item in self:
                if item.get_attr("entity_id") == id_:
                    return item
        raise ValueError("No item with ID {} found in {}".format(
            id_, self.name
        ))

    def get_by_pos(self, pos):
        if not self.group:
            raise ValueError

        # Using low level interface to specify iteration order
        name, _ = self.group.id.links.iterate(lambda n: n,
                                              idx_type=h5py.h5.INDEX_CRT_ORDER,
                                              order=h5py.h5.ITER_INC,
                                              idx=pos)
        return self.get_by_name(name)

    def delete(self, id_or_name):
        if util.is_uuid(id_or_name):
            name = self.get_by_id_or_name(id_or_name).name
        else:
            name = id_or_name
        try:
            del self.group[name]
        except Exception:
            raise ValueError("Error deleting {} from {}".format(name,
                                                                self.name))
        # Delete if empty and non-root container
        groupdepth = len(self.group.name.split("/")) - 1
        if not len(self.group) and groupdepth > 1:
            del self.parent.group[self.name]
            # del self.group
            self.group = None

    def set_attr(self, name, value):
        self._create_h5obj()
        if value is None:
            if name in self.group.attrs:
                del self.group.attrs[name]
        else:
            self.group.attrs[name] = value

    def get_attr(self, name):
        if self.group is None:
            return None
        attr = self.group.attrs.get(name)
        if isinstance(attr, bytes):
            attr = attr.decode()
        return attr

    def find_children(self, filtr=None, limit=None):

        result = []
        start_depth = len(self.group.name.split("/"))

        def match(name, obj):
            curdepth = name.split("/")
            if limit is not None and curdepth == start_depth + limit:
                return None

            h5grp = H5Group.create_from_h5obj(obj)
            if filtr(h5grp):
                result.append(h5grp)

        self.group.visititems(match)
        return result

    @property
    def file(self):
        """
        An H5Group object which represents the file root.

        :return: H5Group at '/'
        """
        return H5Group(self.group.file, "/", create=False)

    @property
    def h5root(self):
        """
        Returns the H5Group of the Block or top-level Section which contains
        this object. Returns None if requested on the file root '/' or the
        /data or /metadata groups.

        :return: Top level object containing this group (H5Group)
        """
        pathparts = self.group.name.split("/")
        if len(pathparts) == 3:
            return self
        if self.group.name == "/":
            return None
        if len(pathparts) == 2:
            return None

        return self.parent.h5root

    @property
    def root(self):
        """
        Returns the Block or top-level Section which contains this object.
        Returns None if requested on the file root '/' or the /data or
        /metadata groups.

        :return: Top level object containing this group (Block or Section)
        """
        h5root = self.h5root
        if h5root is None:
            return None
        topgroup = self.group.name.split("/")[1]
        if topgroup == "data":
            cls = Block
        elif topgroup == "metadata":
            cls = Section
        else:
            raise InvalidEntity
        return cls(h5root)

    @property
    def parent(self):
        return self.create_from_h5obj(self._parent)

    def __iter__(self):
        for grp in self.group.values():
            yield self.create_from_h5obj(grp)

    def __contains__(self, item):
        if self.group is None:
            return False
        return item in self.group

    def __len__(self):
        if self.group is None:
            return 0
        else:
            return len(self.group)

    def __delitem__(self, key):
        del self.group[key]

    def __str__(self):
        return "<H5Group object: {}>".format(self.group.name)



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)


def sanitizer(name):
    """
    Sanitizes a string supposed to be an entity name. That is,
    invalid characters like slashes are substituted with underscores.

    :param name: A string representing the name.

    :returns: The sanitized name.
    :rtype: str
    """
    return name.replace("/", "_")


def check(name):
    """
    Checks a string whether is needs to be sanitized.

    :param name: The name.

    :returns: True if the name is valid, false otherwise.
    :rtype: bool
    """
    if isinstance(name, bytes):
        name = name.decode()
    return "/" not in name



# Copyright (c) 2015, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from nixio.util.proxy_list import RefProxyList


class DataArrayProxyList(RefProxyList):

    def __init__(self, obj):
        super(DataArrayProxyList, self).__init__(
            obj, "_data_array_count", "_get_data_array_by_id",
            "_get_data_array_by_pos", "_delete_data_array_by_id",
            "_add_data_array_by_id"
        )


class TagProxyList(RefProxyList):

    def __init__(self, obj):
        super(TagProxyList, self).__init__(
            obj, "_tag_count", "_get_tag_by_id",
            "_get_tag_by_pos", "_delete_tag_by_id",
            "_add_tag_by_id"
        )


class MultiTagProxyList(RefProxyList):

    def __init__(self, obj):
        super(MultiTagProxyList, self).__init__(
            obj, "_multi_tag_count", "_get_multi_tag_by_id",
            "_get_multi_tag_by_pos", "_delete_multi_tag_by_id",
            "_add_multi_tag_by_id"
        )


class GroupMixin(object):

    @property
    def data_arrays(self):
        """
        A property containing all data arrays referenced by the group.
        Referenced data arrays can be obtained by index or their id. References
        can be removed from the list, removing a referenced DataArray will not
        remove it from the file. New references can be added using the append
        method of the list.
        This is a read only attribute.

        :type: DataArrayProxyList of DataArray
        """
        if not hasattr(self, "_data_arrays"):
            setattr(self, "_data_arrays", DataArrayProxyList(self))
        return self._data_arrays

    @property
    def tags(self):
        """
        A property containing all tags referenced by the group. Tags can be
        obtained by index or their id. Tags can be removed from the list,
        removing a referenced Tag will not remove it from the file.
        New Tags can be added using the append method of the list.
        This is a read only attribute.

        :type: TagProxyList of Tags
        """
        if not hasattr(self, "_tags"):
            setattr(self, "_tags", TagProxyList(self))
        return self._tags

    @property
    def multi_tags(self):
        """
        A property containing all MultiTags referenced by the group. MultiTags
        can be obtained by index or their id. Tags can be removed from the
        list, removing a referenced MultiTag will not remove it from the file.
        New MultiTags can be added using the append method of the list.
        This is a read only attribute.

        :type: MultiTagProxyList of MultiTags
        """
        if not hasattr(self, "_multi_tags"):
            setattr(self, "_multi_tags", MultiTagProxyList(self))
        return self._multi_tags



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function

import nixio as nix


file_name = 'file_create_example.h5'

# create a new file overwriting any existing content
file = nix.File.open(file_name, nix.FileMode.Overwrite)
print(file.format, file.version, file.created_at)

# close file
file.close()

# re-open file for read-only access
file  = nix.File.open(file_name, nix.FileMode.ReadOnly)

# this command will fail putting out HDF5 Errors
file.create_block("test block", "test")

file.close()

# re-open for read-write access
file = nix.File.open(file_name, nix.FileMode.ReadWrite)

# the following command now works fine
file.create_block("test block", "test")

file.close()









# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

from ..multi_tag import MultiTagMixin
from .tag import BaseTag
from .data_array import DataArray
from .data_view import DataView
from ..link_type import LinkType
from .exceptions import (OutOfBounds, IncompatibleDimensions,
                         UninitializedEntity)


class MultiTag(BaseTag, MultiTagMixin):

    def __init__(self, nixparent, h5group):
        super(MultiTag, self).__init__(nixparent, h5group)

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_, positions):
        newentity = super(MultiTag, cls)._create_new(nixparent, h5parent,
                                                     name, type_)
        newentity.positions = positions
        return newentity

    @property
    def positions(self):
        """
        The positions defined by the tag. This is a read-write property.

        :type: DataArray
        """
        return DataArray(self._parent, self._h5group.open_group("positions"))

    @positions.setter
    def positions(self, da):
        if da is None:
            raise TypeError("MultiTag.positions cannot be None.")
        if "positions" in self._h5group:
            del self._h5group["positions"]
        self._h5group.create_link(da, "positions")
        self.force_updated_at()

    @property
    def extents(self):
        """
        The extents defined by the tag. This is an optional read-write
        property and may be set to None.

        :type: DataArray or None
        """
        if "extents" in self._h5group:
            return DataArray(self._parent, self._h5group.open_group("extents"))
        else:
            return None

    @extents.setter
    def extents(self, da):
        if da is None:
            del self._h5group["extents"]
        else:
            self._h5group.create_link(da, "extents")

    def _get_offset_and_count(self, data, index):
        offsets = []
        counts = []
        positions = self.positions
        extents = self.extents

        pos_size = positions.data_extent if positions else tuple()
        ext_size = extents.data_extent if extents else tuple()

        if not positions or index >= pos_size[0]:
            raise OutOfBounds("Index out of bounds of positions!")

        if extents and index >= ext_size[0]:
            raise OutOfBounds("Index out of bounds of extents!")

        if len(pos_size) == 1 and len(data.dimensions) != 1:
            raise IncompatibleDimensions(
                "Number of dimensions in positions does not match "
                "dimensionality of data",
                "MultiTag._get_offset_and_count"
            )

        if len(pos_size) > 1 and pos_size[1] > len(data.dimensions):
            raise IncompatibleDimensions(
                "Number of dimensions in positions does not match "
                "dimensionality of data",
                "MultiTag._get_offset_and_count"
            )

        if (extents and len(ext_size) > 1 and
                ext_size[1] > len(data.dimensions)):
            raise IncompatibleDimensions(
                "Number of dimensions in extents does not match "
                "dimensionality of data",
                "MultiTag._get_offset_and_count"
            )

        offset = positions[index, 0:len(data.dimensions)]
        units = self.units
        for idx in range(offset.size):
            dim = data.dimensions[idx]
            unit = None
            if idx <= len(units) and len(units):
                unit = units[idx]
            offsets.append(self._pos_to_idx(offset.item(idx), unit, dim))

        if extents:
            extent = extents[index, 0:len(data.dimensions)]
            for idx in range(extent.size):
                dim = data.dimensions[idx]
                unit = None
                if idx <= len(units) and len(units):
                    unit = units[idx]
                c = self._pos_to_idx(offset.item(idx) + extent.item(idx),
                                     unit, dim) - offsets[idx]
                counts.append(c if c > 1 else 1)
        else:
            counts = [1]*len(data.dimensions)

        return offsets, counts

    def retrieve_data(self, posidx, refidx):
        references = self.references
        positions = self.positions
        extents = self.extents
        if len(references) == 0:
            raise OutOfBounds("There are no references in this multitag!")

        if (posidx >= positions.data_extent[0] or
                extents and posidx >= extents.data_extent[0]):
            raise OutOfBounds("Index out of bounds of positions or extents!")

        ref = references[refidx]
        dimcount = len(ref.dimensions)
        if len(positions.data_extent) == 1 and dimcount != 1:
            raise IncompatibleDimensions(
                "Number of dimensions in position or extent do not match "
                "dimensionality of data",
                "MultiTag.retrieve_data")
        if len(positions.data_extent) > 1:
            if (positions.data_extent[1] > dimcount or
                    extents and extents.data_extent[1] > dimcount):
                raise IncompatibleDimensions(
                    "Number of dimensions in position or extent do not match "
                    "dimensionality of data",
                    "MultiTag.retrieve_data")
        offset, count = self._get_offset_and_count(ref, posidx)

        if not self._position_and_extent_in_data(ref, offset, count):
            raise OutOfBounds("References data slice out of the extent of the "
                              "DataArray!")
        return DataView(ref, count, offset)

    def retrieve_feature_data(self, posidx, featidx):
        if self._feature_count() == 0:
            raise OutOfBounds(
                "There are no features associated with this tag!"
            )

        try:
            feat = self.features[featidx]
        except KeyError:
            feat = None
            for f in self.features:
                if f.data.name == featidx or f.data.id == featidx:
                    feat = f
                    break
            if feat is None:
                raise
        da = feat.data
        if da is None:
            raise UninitializedEntity()
        if feat.link_type == LinkType.Tagged:
            offset, count = self._get_offset_and_count(da, posidx)
            if not self._position_and_extent_in_data(da, offset, count):
                raise OutOfBounds("Requested data slice out of the extent "
                                  "of the Feature!")
            return DataView(da, count, offset)
        elif feat.link_type == LinkType.Indexed:
            if posidx > da.data_extent[0]:
                raise OutOfBounds("Position is larger than the data stored "
                                  "in the Feature!")
            offset = [0] * len(da.data_extent)
            offset[0] = posidx
            count = list(da.data_extent)
            count[0] = 1

            if not self._position_and_extent_in_data(da, offset, count):
                OutOfBounds("Requested data slice out of the extent of the "
                            "Feature!")
            return DataView(da, count, offset)
        # For untagged return the full data
        count = da.data_extent
        offset = (0,) * len(count)
        return DataView(da, count, offset)



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from __future__ import (absolute_import, division, print_function)

from . import util


class Entity(object):

    def __init__(self, nixparent, h5group):
        util.check_entity_id(h5group.get_attr("entity_id"))
        self._h5group = h5group
        self._parent = nixparent

    @classmethod
    def _create_new(cls, nixparent, h5parent):
        id_ = util.create_id()
        h5group = h5parent.open_group(id_)
        h5group.set_attr("entity_id", id_)
        newentity = cls(nixparent, h5group)
        newentity.force_created_at()
        newentity.force_updated_at()
        return newentity

    @property
    def id(self):
        """
        A property providing the ID of the Entity. The id is generated
        automatically, therefore the property is read-only.

        :rtype: str
        """
        return self._h5group.get_attr("entity_id")

    @property
    def created_at(self):
        """
        The creation time of the entity. This is a read-only property.
        Use `force_created_at` in order to change the creation time.

        :rtype: int
        """
        return util.str_to_time(self._h5group.get_attr("created_at"))

    def force_created_at(self, t=None):
        """
        Sets the creation time `created_at` to the given time
        (default: current time).

        :param t: The time to set.
        :type t: int
        """
        if t is None:
            t = util.now_int()
        else:
            util.check_attr_type(t, int)
        self._h5group.set_attr("created_at", util.time_to_str(t))

    @property
    def updated_at(self):
        """
        The time of the last update of the entity. This is a read-only
        property. Use `force_updated_at` in order to change the update
        time.

        :rtype: int
        """
        return util.str_to_time(self._h5group.get_attr("updated_at"))

    def force_updated_at(self, t=None):
        """
        Sets the update time `updated_at` to the given time.
        (default: current time)

        :param t: The time to set.
        :type t: int
        """
        if t is None:
            t = util.now_int()
        else:
            util.check_attr_type(t, int)
        self._h5group.set_attr("updated_at", util.time_to_str(t))

    def __eq__(self, other):
        if hasattr(other, "id"):
            return self.id == other.id
        else:
            return False

    def __hash__(self):
        return hash(self.id)


class NamedEntity(Entity):

    def __init__(self, nixparent, h5group):
        try:
            util.check_entity_name_and_type(h5group.get_attr("name"),
                                            h5group.get_attr("type"))
            util.check_entity_id(h5group.get_attr("entity_id"))
        except ValueError:
            ValueError("Invalid NIX object found in file.")
        super(NamedEntity, self).__init__(nixparent, h5group)

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_):
        util.check_entity_name_and_type(name, type_)
        h5group = h5parent.open_group(name)
        h5group.set_attr("name", name)
        h5group.set_attr("type", type_)
        h5group.set_attr("entity_id", util.create_id())
        newentity = cls(nixparent, h5group)
        newentity.force_created_at()
        newentity.force_updated_at()
        return newentity

    @property
    def name(self):
        """
        The name of an entity. The name serves as a human readable
        identifier. This is a read-only property; entities cannot be
        renamed.

        :type: str
        """
        return self._h5group.get_attr("name")

    @property
    def type(self):
        """
        The type of the entity. The type is used in order to add semantic
        meaning to the entity. This is a read-write property, but it can't
        be set to None.

        :type: str
        """
        return self._h5group.get_attr("type")

    @type.setter
    def type(self, t):
        if t is None:
            raise AttributeError("type can't be None")
        util.check_attr_type(t, str)
        self._h5group.set_attr("type", t)

    @property
    def definition(self):
        """
        The definition of the entity. The definition can contain a textual
        description of the entity. This is an optional read-write
        property, and can be None if no definition is available.

        :type: str
        """
        return self._h5group.get_attr("definition")

    @definition.setter
    def definition(self, d):
        util.check_attr_type(d, str)
        self._h5group.set_attr("definition", d)

    def __str__(self):
        return "{}: {{name = {}, type = {}, id = {}}}".format(
            type(self).__name__, self.name, self.type, self.id
        )

    def __repr__(self):
        return self.__str__()



# Copyright (c) 2016, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.
from numbers import Number

from .entity_with_sources import EntityWithSources
from ..data_array import DataArrayMixin, DataSetMixin
from ..value import DataType
from .dimensions import (SampledDimension, RangeDimension, SetDimension,
                         DimensionType)
from . import util

from .exceptions import InvalidUnit


class DataSet(DataSetMixin):

    def _write_data(self, data, count, offset):
        dataset = self._h5group.get_dataset("data")
        dataset.write_data(data, count, offset)

    def _read_data(self, data, count, offset):
        dataset = self._h5group.get_dataset("data")
        dataset.read_data(data, count, offset)

    @property
    def data_extent(self):
        """
        The size of the data.

        :type: set of int
        """
        dataset = self._h5group.get_dataset("data")
        return dataset.shape

    @data_extent.setter
    def data_extent(self, extent):
        dataset = self._h5group.get_dataset("data")
        dataset.shape = extent

    @property
    def data_type(self):
        """
        The data type of the data stored in the DataArray. This is a read only
        property.

        :type: DataType
        """
        return self._get_dtype()

    def _get_dtype(self):
        dataset = self._h5group.get_dataset("data")
        return dataset.dtype


class DataArray(EntityWithSources, DataSet, DataArrayMixin):

    def __init__(self, nixparent, h5group):
        super(DataArray, self).__init__(nixparent, h5group)

    @classmethod
    def _create_new(cls, nixparent, h5parent, name, type_, data_type, shape):
        newentity = super(DataArray, cls)._create_new(nixparent, h5parent,
                                                      name, type_)
        newentity._h5group.create_dataset("data", shape, data_type)
        return newentity

    def _read_data(self, data, count, offset):
        coeff = self.polynom_coefficients
        origin = self.expansion_origin
        if len(coeff) or origin:
            if not origin:
                origin = 0.0

            super(DataArray, self)._read_data(data, count, offset)
            util.apply_polynomial(coeff, origin, data)
        else:
            super(DataArray, self)._read_data(data, count, offset)

    def append_set_dimension(self):
        """
        Append a new SetDimension to the list of existing dimension
        descriptors.

        :returns: The newly created SetDimension.
        :rtype: SetDimension
        """
        dimgroup = self._h5group.open_group("dimensions")
        index = len(dimgroup) + 1
        return SetDimension._create_new(dimgroup, index)

    def append_sampled_dimension(self, sampling_interval):
        """
        Append a new SampledDimension to the list of existing dimension
        descriptors.

        :param sampling_interval: The sampling interval of the SetDimension
                                  to create.
        :type sampling_interval: float

        :returns: The newly created SampledDimension.
        :rtype: SampledDimension
        """
        dimgroup = self._h5group.open_group("dimensions")
        index = len(dimgroup) + 1
        return SampledDimension._create_new(dimgroup, index, sampling_interval)

    def append_range_dimension(self, ticks):
        """
        Append a new RangeDimension to the list of existing dimension
        descriptors.

        :param ticks: The ticks of the RangeDimension to create.
        :type ticks: list of float

        :returns: The newly created RangeDimension.
        :rtype: RangeDimension
        """
        dimgroup = self._h5group.open_group("dimensions")
        index = len(dimgroup) + 1
        return RangeDimension._create_new(dimgroup, index, ticks)

    def append_alias_range_dimension(self):
        """
        Append a new RangeDimension that uses the data stored in this
        DataArray as ticks. This works only(!) if the DataArray is 1-D and
        the stored data is numeric. A ValueError will be raised otherwise.

        :returns: The created dimension descriptor.
        :rtype: RangeDimension
        """
        if (len(self.data_extent) > 1 or
                not DataType.is_numeric_dtype(self.dtype)):
            raise ValueError("AliasRangeDimensions only allowed for 1D "
                             "numeric DataArrays.")
        if self._dimension_count() > 0:
            raise ValueError("Cannot append additional alias dimension. "
                             "There must only be one!")
        dimgroup = self._h5group.open_group("dimensions")
        # check if existing unit is SI
        if self.unit:
            u = self.unit
            if not (util.units.is_si(u) or util.units.is_compound(u)):
                raise InvalidUnit(
                    "AliasRangeDimensions are only allowed when SI or "
                    "composites of SI units are used. "
                    "Current SI unit is {}".format(u),
                    "DataArray.append_alias_range_dimension"
                )
        return RangeDimension._create_new_alias(dimgroup, 1, self)

    def delete_dimensions(self):
        """
        Delete all the dimension descriptors for this DataArray.
        """
        dimgroup = self._h5group.open_group("dimensions")
        ndims = len(dimgroup)
        for idx in range(ndims):
            del dimgroup[str(idx+1)]
        return True

    def _dimension_count(self):
        return len(self._h5group.open_group("dimensions"))

    def _get_dimension_by_pos(self, index):
        h5dim = self._h5group.open_group("dimensions").open_group(str(index))
        dimtype = h5dim.get_attr("dimension_type")
        if dimtype == DimensionType.Sample:
            return SampledDimension(h5dim, index)
        elif dimtype == DimensionType.Range:
            return RangeDimension(h5dim, index)
        elif dimtype == DimensionType.Set:
            return SetDimension(h5dim, index)
        else:
            raise TypeError("Invalid Dimension object in file.")

    @property
    def dtype(self):
        """
        The data type of the data stored in the DataArray.
        This is a read only property.

        :return: DataType
        """
        return self._h5group.group["data"].dtype

    @property
    def polynom_coefficients(self):
        """
        The polynomial coefficients for the calibration. By default this is
        set to a {0.0, 1.0} for a linear calibration with zero offset.
        This is a read-write property and can be set to None

        :type: list of float
        """
        return tuple(self._h5group.get_data("polynom_coefficients"))

    @polynom_coefficients.setter
    def polynom_coefficients(self, coeff):
        if not coeff:
            if self._h5group.has_data("polynom_coefficients"):
                del self._h5group["polynom_coefficients"]
        else:
            dtype = DataType.Double
            self._h5group.write_data("polynom_coefficients", coeff, dtype)

    @property
    def expansion_origin(self):
        """
        The expansion origin of the calibration polynomial.
        This is a read-write property and can be set to None.
        The default value is 0.

        :type: float
        """
        return self._h5group.get_attr("expansion_origin")

    @expansion_origin.setter
    def expansion_origin(self, eo):
        util.check_attr_type(eo, Number)
        self._h5group.set_attr("expansion_origin", eo)

    @property
    def label(self):
        """
        The label of the DataArray. The label corresponds to the label of the
        x-axis of a plot. This is a read-write property and can be set to
        None.

        :type: str
        """
        return self._h5group.get_attr("label")

    @label.setter
    def label(self, l):
        util.check_attr_type(l, str)
        self._h5group.set_attr("label", l)

    @property
    def unit(self):
        """
        The unit of the values stored in the DataArray. This is a read-write
        property and can be set to None.

        :type: str
        """
        return self._h5group.get_attr("unit")

    @unit.setter
    def unit(self, u):
        if u:
            u = util.units.sanitizer(u)
        if u == "":
            u = None
        util.check_attr_type(u, str)
        if (self._dimension_count() == 1 and
                self.dimensions[0].dimension_type == DimensionType.Range and
                self.dimensions[0].is_alias and u is not None):
            if not (util.units.is_si(u) or util.units.is_compound(u)):
                raise InvalidUnit(
                    "[{}]: Non-SI units are not allowed if the DataArray "
                    "has an AliasRangeDimension.".format(u),
                    "DataArray.unit"
                )
        self._h5group.set_attr("unit", u)



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

import functools

import nixio.util.find as finders
from nixio.util.proxy_list import ProxyList
from nixio.value import Value

from nixio.file import SectionProxyList

from operator import attrgetter

try:
    from sys import maxint
except:
    from sys import maxsize as maxint


class S(object):
    def __init__(self, section_type, section=None):
        self.section_type = section_type
        self.section = section

    def __setitem__(self, key, value):
        self.section[key] = value

    def __setattr__(self, key, value):
        if key in ['section_type', 'section']:
            object.__setattr__(self, key, value)
        else:
            setattr(self.section, key, value)

    def __getattribute__(self, item):
        if item in ['section_type', 'section']:
            return object.__getattribute__(self, item)
        else:
            return getattr(self.section, item)


class PropertyProxyList(ProxyList):

    def __init__(self, obj):
        super(PropertyProxyList, self).__init__(obj,
                                                "_property_count",
                                                "_get_property_by_id_or_name",
                                                "_get_property_by_pos",
                                                "_delete_property_by_id")


class SectionMixin(object):

    def find_sections(self, filtr=lambda _: True, limit=None):
        """
        Get all child sections recursively.
        This method traverses the trees of all sections. The traversal is
        accomplished via breadth first and can be limited in depth.
        On each node or section a filter is applied.
        If the filter returns true the respective section will be added to the
        result list.  By default a filter is used that accepts all sections.

        :param filtr: A filter function
        :type filtr:  function
        :param limit: The maximum depth of traversal
        :type limit:  int

        :returns: A list containing the matching sections.
        :rtype: list of Section
        """
        if limit is None:
            limit = maxint
        return finders._find_sections(self, filtr, limit)

    def find_related(self, filtr=lambda _: True):
        """
        Get all related sections of this section.

        The result can be filtered. On each related section a filter is
        applied.  If the filter returns true the respective section will be
        added to the result list. By default a filter is used that accepts all
        sections.

        :param filtr: A filter function
        :type filtr:  function

        :returns: A list containing the matching related sections.
        :rtype: list of Section
        """
        result = []
        if self.parent is not None:
            result = finders._find_sections(self.parent, filtr, 1)
        if self in result:
            del result[result.index(self)]
        result += finders._find_sections(self, filtr, 1)
        return result

    @property
    def sections(self):
        """
        A property providing all child sections of a section. Child sections
        can be accessed by index or by their id. Sections can also be deleted:
        if a section is deleted, all its properties and child sections are
        removed from the file too. Adding new sections is achieved using the
        create_section method.
        This is a read-only attribute.

        :type: ProxyList of Section
        """
        if not hasattr(self, "_sections"):
            setattr(self, "_sections", SectionProxyList(self))
        return self._sections

    def __eq__(self, other):
        if hasattr(other, "id"):
            return self.id == other.id
        else:
            return False

    def __hash__(self):
        """
        overwriting method __eq__ blocks inheritance of __hash__ in Python 3
        hash has to be either explicitly inherited from parent class,
        implemented or escaped
        """
        return hash(self.id)

    def __len__(self):
        return len(self.props)

    def __getitem__(self, key):

        if key not in self.props and key in self.sections:
            return self.sections[key]

        prop = self.props[key]
        values = list(map(attrgetter('value'), prop.values))
        if len(values) == 1:
            values = values[0]
        return values

    def __delitem__(self, key):
        del self.props[key]

    def __setitem__(self, key, data):

        if isinstance(data, S):
            data.section = self.create_section(key, data.section_type)
            return

        if not isinstance(data, list):
            data = [data]

        val = list(map(lambda x: x if isinstance(x, Value) else Value(x),
                       data))
        dtypes = functools.reduce(
            lambda x, y: x if y.data_type in x else x + [y.data_type],
            val, [val[0].data_type]
        )
        if len(dtypes) > 1:
            raise ValueError('Not all input values are of the same type')

        if key not in self.props:
            prop = self.create_property(key, dtypes[0])
        else:
            prop = self.props[key]
        prop.values = val

    def __iter__(self):
        for name, item in self.items():
            yield item

    def items(self):
        for p in self.props:
            yield (p.name, p)
        for s in self.sections:
            yield (s.name, s)

    def __contains__(self, key):
        return key in self.props or key in self.sections

    @property
    def props(self):
        """
        A property containing all Property entities associated with the
        section.  Properties can be accessed by index of via their id.
        Properties can be deleted from the list. Adding new properties is done
        using the create_property method.
        This is a read-only attribute.

        :type: ProxyList of Property
        """
        if not hasattr(self, "_properties_proxy"):
            setattr(self, "_properties_proxy", PropertyProxyList(self))
        return self._properties_proxy

    def _get_property_by_id_or_name(self, ident):
        """
        Helper method that gets a property by id or name
        """
        p = self._get_property_by_id(ident)
        if p is None and self.has_property_by_name(ident):
            p = self.get_property_by_name(ident)
        return p



# Copyright (c) 2014, German Neuroinformatics Node (G-Node)
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted under the terms of the BSD License. See
# LICENSE file in the root of the Project.

from nixio.info import RELEASE, COPYRIGHT, BRIEF

# general config
extensions              = ['sphinx.ext.autodoc', 'sphinx.ext.intersphinx']
source_suffix           = '.rst'
master_doc              = 'index'
project                 = BRIEF
copyright               = COPYRIGHT
release                 = RELEASE
exclude_patterns        = []
pygments_style          = 'sphinx'

# html options
htmlhelp_basename       = 'nixio'
try:
    import alabaster

    html_theme_path     = [alabaster.get_path()]
    html_static_path    = ['_static']
    extensions          += ['alabaster']
    html_theme          = 'alabaster'

    html_sidebars       = {
        '**': [
                'about.html', 'navigation.html', 'searchbox.html',
            ]
    }

    html_theme_options  = {
        'logo'          : 'logo.png',
        'github_user'   : 'G-Node',
        'github_repo'   : 'nixpy',
        'github_button' : True,
        'github_count'  : False,
        'travis_button' : True,
        'link'          : '#456BA8'
    }

except ImportError:
    html_theme          = 'default'

# intersphinx configuration
intersphinx_mapping     = {
    'http://docs.python.org/2.7' : None,
    'http://docs.scipy.org/doc/numpy': None
}



from __future__ import print_function
from __future__ import division

import os
import re
import sys
from operator import itemgetter


try:
    from commands import getstatusoutput
except ImportError:
    from subprocess import getstatusoutput


class BoostPyLib(object):
    matcher = re.compile(r'^libboost_python(?:-py)?(\d{0,2})((?:-\w+[^-])*)\.(?:so|dylib|lib)$')

    def __init__(self, path, name, major, minor, ismt):
        self.path = path
        self.filename = name
        self.major = major
        self.minor = minor
        self.threadsafe = ismt

    def match(self, major, minor, threadsafe, ignore_threading=False):
        return self.major == major and self.minor == minor and (ignore_threading or self.threadsafe == threadsafe)

    @property
    def library_name(self):
        name, ext = os.path.splitext(self.filename)
        return name if ext == ".lib" else name[3:]

    @property
    def link_directive(self):
        name, ext = os.path.splitext(self.filename)
        if ext == '.lib':
            return ["/DEFAULTLIB:" + self.filename, '/LIBPATH:' + self.path]
        else:
            return ['-l' + name[3:]]

    def __str__(self):
        major = str(self.major or '?')
        minor = str(self.minor or '?')
        mt = 'threadsafe' if self.threadsafe else ''
        return "%s [%s.%s] (%s) " % (self.filename, major, minor, mt)

    @staticmethod
    def library_search_dirs(additional_dirs=None):
        dirs = additional_dirs or []
        cc = os.getenv("CC", "cc")
        cmd = "%s --print-search-dirs | grep ^libraries" % cc
        res, out = getstatusoutput(cmd)
        if res != 0:
            print("WARNING: \"%s\" failed with %d" % (cmd, res), file=sys.stderr)
            return dirs

        start = out.find("=") + 1
        paths = out[start:].split(":")
        ordered_set = {k: idx for idx, k in enumerate(paths)}
        ordered_set.update({d: -1*(len(dirs) - i) for i, d in enumerate(dirs)})
        return list(map(itemgetter(0), (sorted(ordered_set.items(), key=itemgetter(1)))))

    @classmethod
    def list_in_dirs(cls, dirs):
        return list(cls.iterate_libs_in_dirs(dirs))

    @classmethod
    def iterate_libs_in_dirs(cls, dirs):
        for d in dirs:
            if not os.path.isdir(d):
                continue
            for f in os.listdir(d):
                l = cls.make_from_path(f)
                if l is not None:
                    yield l

    @classmethod
    def make_from_path(cls, path):
        filename = os.path.basename(path)
        dirname = os.path.dirname(path)
        m = cls.matcher.match(filename)
        if m is None:
            return m
        ismt = '-mt' in m.group(2)
        major, minor = cls.parse_version(m.group(1))
        return cls(dirname, filename, major, minor, ismt)

    @classmethod
    def find_lib_for_current_python(cls, libs):
        v_major, v_minor = sys.version_info[:2]
        match = cls.find_lib_with_version(libs, (v_major, v_minor), unknown_is_match=v_major == 2)
        return match

    @staticmethod
    def find_lib_with_version(available, version, unknown_is_match=False):
        major, minor = version[0], version[1]
        criteria = [(major, minor), (major, None)] + ([(None, None)] if unknown_is_match else [])
        for c in [[c[0], c[1], mt] for c in criteria for mt in [True, False]]:
            res = list(filter(lambda x: x.match(*c), available))
            if len(res) > 0:
                return res[0]
        return None

    @staticmethod
    def parse_version(input):
        major, minor = None, None
        try:
            major = int(input[0])
            minor = int(input[1])
        except (IndexError, ValueError):
            pass
        return major, minor

    @classmethod
    def selftest(cls):
        survey = [
            ('libboost_python.so', (None, None, False)),
            ('libboost_python3.so', (3, None, False)),
            ('libboost_python.dylib', (None, None, False)),
            ('libboost_python3.dylib', (3, None, False)),
            ('libboost_python-py27.so', (2, 7, False)),
            ('libboost_python-py27.so.1.55.0', None),
            ('libboost_python-py34.so', (3, 4, False)),
            ('libboost_python-py35.so', (3, 5, False)),
            ('libboost_python.lib', (None, None, False)),
            ('libboost_python-mt.dylib', (None, None, True)),
            ('libboost_python-vc120-mt-1_57.lib', (None, None, True)),
            ('libboost_python-vc120-mt-1_61.lib', (None, None, True)),
            ('/usr/lib/libboost_python.a', None)]

        try:
            ml = max(map(lambda s: len(s[0]), survey)) + 3
            for s, c in survey:
                o = cls.make_from_path(s)
                sys.stderr.write('Checking: ' + str(o and str(o.match(*c))) + s)
                assert(c is not None or o is None)
                assert(o is not None or c is None)
                assert(o is None or o.match(*c))
                sys.stderr.write(' '*(ml - len(s)) + 'ok\n')
        except AssertionError as e:
            print("Self-test failed:", e, file=sys.stderr)
            return -1
        return 0


if __name__ == '__main__':
    import argparse
    default_path = os.getenv('NIX_LIBDIR', '/usr/local/lib')
    parser = argparse.ArgumentParser('NixPy - find boost python libraries')
    parser.add_argument('--path', default=default_path, help='Path to search libraries in [%s]' % default_path)
    parser.add_argument('--selftest', default=False, action='store_true', help='Perform a self-test')
    args = parser.parse_args()

    v_major, v_minor = sys.version_info[:2]

    print('Python version: [%d.%d]' % (v_major, v_minor), file=sys.stderr)
    if args.selftest:
        ret = BoostPyLib.selftest()
        sys.exit(ret)

    libs = BoostPyLib.list_in_dirs([default_path])
    for lib in libs:
        print(' %s (ld: %s)' % (str(lib), lib.link_directive), file=sys.stderr)
        linked = BoostPyLib.find_lib_with_version(libs, (v_major, v_minor), unknown_is_match=v_major == 2)
        if linked is not None:
            print(' match: ' + str(linked), file=sys.stderr)
            print(' linker command: ' + str(linked.link_directive))
            print(' library name: ' + linked.library_name)
        else:
            print('No usable boost python lib found! :(', file=sys.stderr)
            sys.exit(-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
 Copyright © 2014 German Neuroinformatics Node (G-Node)

 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted under the terms of the BSD License. See
 LICENSE file in the root of the Project.

 Author: Jan Grewe <jan.grewe@g-node.org>

 This tutorial shows how regulary sampled data is stored in nix-files.
 See https://github.com/G-node/nix/wiki for more information.
"""

import nixio as nix
import numpy as np
import matplotlib.pylab as plt


def create_sinewave(duration=1, freq=10, stepsize=0.01):
    x = np.arange(0, duration*2*np.pi, stepsize)
    y = np.sin(freq*x)
    return x, y


def plot_data(data_array):
    x_axis = data_array.dimensions[0]
    x = x_axis.axis(data_array.data.shape[0])
    y = data_array.data[:]
    plt.plot(x, y, marker=".", markersize=5)
    plt.xlabel(x_axis.label + " [" + x_axis.unit + "]")
    plt.ylabel(data_array.label + " [" + data_array.unit + "]")
    plt.title(data_array.name)
    plt.xlim(0, np.max(x))
    plt.ylim((1.1 * np.min(y), 1.1 * np.max(y)))
    plt.show()


if __name__ == "__main__":
    # fake some data
    duration = 1.
    frequency = 5
    stepsize = 0.02
    x, y = create_sinewave(duration, frequency, stepsize)

    # create a new file overwriting any existing content
    file_name = 'regular_data_example.h5'
    file = nix.File.open(file_name, nix.FileMode.Overwrite)

    # create a 'Block' that represents a grouping object. Here, the recording session.
    # it gets a name and a type
    block = file.create_block("block name", "nix.session")

    # create a 'DataArray' to take the sinewave, add some information about the signal
    data = block.create_data_array("sinewave", "nix.regular_sampled", data=y)
    data.unit = "mV"
    data.label = "voltage"
    # add a descriptor for the xaxis
    dim = data.append_sampled_dimension(stepsize)
    dim.unit = "s"
    dim.label = "time"
    dim.offset = 0.0 # optional

    # let's plot the data from the stored information
    plot_data(data)
    file.close()



#!/usr/bin/env python

from __future__ import absolute_import, division, print_function

__author__ = 'gicmo'

try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

# python version dependent import of getstatusoutput
try:
    from commands import getstatusoutput
except ImportError:
    from subprocess import getstatusoutput

try:
    import numpy as np
    np_inc_dir = np.get_include()
except ImportError:
    np_inc_dir = ""

import sys
import os
import json

from scripts.findboost import BoostPyLib
from scripts.checknix import check_nix


# Replaced StandardError with Exception since StandardError is removed in Py3.x
class PackageNotFoundError(Exception):
    pass


with open('README.rst') as f:
    description_text = f.read()

with open('LICENSE') as f:
    license_text = f.read()


is_win = os.name == 'nt'

# load info from nixio/info.json
with open(os.path.join("nixio", "info.json")) as infofile:
    infodict = json.load(infofile)

VERSION = infodict["VERSION"]
AUTHOR = infodict["AUTHOR"]
CONTACT = infodict["CONTACT"]
BRIEF = infodict["BRIEF"]
HOMEPAGE = infodict["HOMEPAGE"]


if "dev" in VERSION:
    if is_win:
        colorcodes = ("", "")
    else:
        colorcodes = ("\033[93m", "\033[0m")
    sys.stderr.write("{}WARNING: You are building a development version "
                     "of nixpy.{}\n".format(*colorcodes))


def pkg_config(*packages, **kw):
    flag_map = {'-I': 'include_dirs', '-L': 'library_dirs', '-l': 'libraries'}

    ignore_error = 'ignore_error' in kw
    if ignore_error:
        del kw['ignore_error']

    pkg_string = ' '.join(packages)
    status, out = getstatusoutput("pkg-config --libs --cflags " + pkg_string)
    if status != 0:
        err_str = 'Some packages were not found: %s\n\t%s' % (
            pkg_string, out.replace("\n", "\n\t")
        )
        if ignore_error:
            sys.stderr.write('WARNING: {}\n'.format(err_str))
            out = ''
        else:
            raise PackageNotFoundError(err_str)

    for token in out.split():
        kw.setdefault(flag_map.get(token[:2]), []).append(token[2:])

    # remove duplicated
    # could lead to increased memory usage and computation time.
    for k, v in kw.items():
        del kw[k]
        kw[k] = list(set(v))

    return kw


def get_wheel_data():
    data = []
    bin = os.environ.get('NIXPY_WHEEL_BINARIES', '')
    if bin and os.path.isdir(bin):
        data.append(
            ('share/nixio/bin',
             [os.path.join(bin, f) for f in os.listdir(bin)]))
    return data


nix_inc_dir = os.getenv('NIX_INCDIR', '/usr/local/include')
nix_lib_dir = os.getenv('NIX_LIBDIR', '/usr/local/lib')
nix_lib = 'nix'

nixpy_sources = [
    'src/core.cc',
    'src/docstrings.cpp',
    'src/PyExceptions.cpp',
    'src/PyBlock.cpp',
    'src/PyFile.cpp',
    'src/PySection.cpp',
    'src/PyProperty.cpp',
    'src/PySource.cpp',
    'src/PyDataArray.cpp',
    'src/PyDataSet.cpp',
    'src/PyDimensions.cpp',
    'src/PyFeature.cpp',
    'src/PyTag.cpp',
    'src/PyMultiTag.cpp',
    'src/PyResult.cpp',
    'src/PyGroup.cpp'
]

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Programming Language :: Python',
    'Programming Language :: Python :: 2.6',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Topic :: Scientific/Engineering'
]

boost_inc_dir = os.getenv('BOOST_INCDIR', '/usr/local/include')
boost_lib_dir = os.getenv('BOOST_LIBDIR', '/usr/local/lib')
lib_dirs = BoostPyLib.library_search_dirs([boost_lib_dir])\
    if not is_win else []
boost_libs = BoostPyLib.list_in_dirs(lib_dirs)
boost_lib = BoostPyLib.find_lib_for_current_python(boost_libs)
library_dirs = [boost_lib_dir, nix_lib_dir]
include_dirs = [boost_inc_dir, nix_inc_dir, np_inc_dir, 'src']
libraries = [nix_lib]
compile_args = ['--std=c++11'] if not is_win else ['/DBOOST_PYTHON_STATIC_LIB',
                                                   '/EHsc']

if "--without-nix" in sys.argv:
    sys.argv.remove("--without-nix")
    with_nix = False
elif "--with-nix" in sys.argv:
    sys.argv.remove("--with-nix")
    with_nix = True
else:
    with_nix = check_nix(library_dirs, include_dirs, compile_args)

if with_nix:
    if boost_lib is None:
        print("Could not find boost python version for {}.{}".format(
            *sys.version_info[0:2]))
        print("Available boost python libs:")
        print("\n".join(map(str, boost_libs)))
        sys.exit(-1)

    libraries.append(boost_lib.library_name)

    extargs = dict(
        extra_compile_args=['-std=c++11']
        if not is_win else ['/DBOOST_PYTHON_STATIC_LIB', '/EHsc'],
        libraries=libraries,
        sources=nixpy_sources,
        runtime_library_dirs=library_dirs if not is_win else None,
    )

    pc = pkg_config("nix", library_dirs=library_dirs,
                    include_dirs=include_dirs, ignore_error=False)
    for k in pc:
        if k in extargs:
            pc[k] = list(set(pc[k] + extargs[k]))
    extargs.update(pc)
    native_ext = Extension('nixio.core', **extargs)
    ext_modules = [native_ext]
else:
    print("Skipping NIX C++ bindings.")
    ext_modules = []

setup(
    name='nixio',
    version=VERSION,
    author=AUTHOR,
    author_email=CONTACT,
    url=HOMEPAGE,
    description=BRIEF,
    long_description=description_text,
    classifiers=classifiers,
    license='BSD',
    ext_modules=ext_modules,
    packages=['nixio', 'nixio.pycore', 'nixio.util',
              'nixio.pycore.util', 'nixio.pycore.exceptions'],
    scripts=[],
    tests_require=['pytest'],
    test_suite='pytest',
    setup_requires=['pytest-runner'],
    install_requires=['numpy', 'h5py'],
    package_data={'nixio': [license_text, description_text]},
    include_package_data=True,
    zip_safe=False,
    data_files=get_wheel_data()
)
